local MAJOR = "StarLibPluginResourceTools-1.0" 
local MINOR = 11

local PluginResourceTools = LibStub:NewLibrary(MAJOR, MINOR)
if not PluginResourceTools then return end

if not PluginResourceTools.__index then
	PluginResourceTools.__index = PluginResourceTools
end

local scriptProfile = GetCVar("scriptProfile") == "1"
local oldtotalcpu, oldtotalmem
local lastcputime, lastmemtime = GetTime(), GetTime()

local oldmem = {}
local oldcpu = {}
local meminit
local cpuinit

local new, del
do
	local pool = setmetatable({}, {__mode = "k"})
	function new(...)
		local obj = next(pool)
		if obj then
			pool[obj] = nil
		else
			obj = {}
		end
		for i = 1, select("#", ...) do
			obj[i] = select(i, ...)
		end
		return obj
	end
	function del(tbl)
		pool[tbl] = true
	end
end

-- @name PluginResourceTools:New
-- @usage PluginResourceTools:New(visitor, environment, name, config) -- 
-- @parma environment This will be the environment when loadstring is callled.
-- @return A new StarPluginResourceTools object
function PluginResourceTools:New(environment)
	
	local obj = {}

	setmetatable(obj, self)
	
	environment.GetCPUUsage = self.GetCPUUsage
	environment.GetMemUsage = self.GetMemUsage
	environment.GetLatency = self.GetLatency
	environment.UpdateCPU = self.UpdateCPU
	environment.UpdateMem = self.UpdateMem
	environment.Update = self.Update
	environment.GetFramerate = self.GetFramerate
	environment.scriptProfile = scriptProfile
	
	return obj	
	
end

-- @name PluginResourceTools:Update
-- @usage PluginResourceTools:Update()
-- @return Nothing
function PluginResourceTools:Update()
	PluginResourceTools:UpdateMem()
	PluginResourceTools:UpdateCPU()
end

-- @name UpdateCPU
-- @usage PluginResourceTools.UpdateCPU()
-- @return Nothing
function PluginResourceTools.UpdateCPU()
	cpuinit = true
	
	UpdateAddOnCPUUsage()
	
	local totalcpu = 0.001
	
	for i=1, GetNumAddOns() do
		totalcpu = totalcpu + (GetAddOnCPUUsage(i) )
	end	
	
	for i = 1, GetNumAddOns() do
	
		local addon = GetAddOnInfo(i) 
		
		if not oldcpu[addon] then 
			oldcpu[addon] = new(-1, -1, -1, -1, -1)
		end
		
		del(oldcpu[addon])
						
		local cpu = GetAddOnCPUUsage(addon)

		local perc = cpu / totalcpu * 100
		local cpudiff = oldcpu[addon] and oldcpu[addon][1] and (cpu - oldcpu[addon][1]) or 0
		local totaldiff = oldtotalcpu and (totalcpu - oldtotalcpu) or 0

		oldcpu[addon][1] = cpu
		oldcpu[addon][2] = cpuperc
		oldcpu[addon][3] = cpudiff
		oldcpu[addon][4] = totalcpu
		oldcpu[addon][5] = totaldiff
	end
	oldtotalcpu = totalcpu	
end

-- @name GetCPUUsage
-- @usage PluginResourceTools.GetCPUUsage(addon)
-- @param addon We'll get the cpu usage for this addon.
-- @param usestarlibs Use this to change the addon name to StarLibs-1.0 if StarLibs-1.0, the addon, is loaded.
-- @return cpu, percent, cpudiff, totalcpu, totaldiff
function PluginResourceTools.GetCPUUsage(addon, usestarlibs)
	if not scriptProfile then return 0, 0, 0, 0, 0 end
	
	assert(type(addon) == "string", "GetCPUUsage: addon name is not a string")
	
	if not cpuinit then return -1, -1, -1, -1, -1 end
	
	if StarLibs and (addon == "StarTip" or addon == "LCD4WoW" or usestarlibs) then
		addon = "StarLibs-1.0"
	end

	if not cpuinit then PluginResourceTools:Update() end
	
	local cpu = oldcpu[addon][1]
	local perc = oldcpu[addon][2]
	local cpudiff = oldcpu[addon][3]
	local totalcpu = oldcpu[addon][4]
	local totaldiff = oldcpu[addon][5]
	
	return cpu, perc, cpudiff, totalcpu, totaldiff
end

-- @name UpdateMem
-- @usage PluginResourceTools.Update()
-- @return Nothing
function PluginResourceTools.UpdateMem() 
	UpdateAddOnMemoryUsage()
	
	meminit = true
	
	if not oldmem then oldmem = {} end
	
	local totalmem = 0.0001
	
	for i=1, GetNumAddOns() do
		totalmem = totalmem + (GetAddOnMemoryUsage(i) )
	end	

	for i = 1, GetNumAddOns() do	
		local addon = GetAddOnInfo(i) 
		
		if not oldmem[addon] then 
			oldmem[addon] = new(-1, -1, -1, -1, -1)
		end
		
		del(oldmem[addon])
				
		local mem = GetAddOnMemoryUsage(addon)
		local perc = mem / totalmem * 100
		local memdiff = oldmem[addon][1] and (mem - oldmem[addon][1]) or 0
		local totaldiff = oldtotalmem and (totalmem - oldtotalmem) or 0
		
				
		oldmem[addon] = new() 
		oldmem[addon][1] = mem
		oldmem[addon][2] = memperc
		oldmem[addon][3] = memdiff
		oldmem[addon][4] = totalcpu
		oldmem[addon][5] = totaldiff
	end
	oldtotalmem = totalmem
end

-- @name GetMemUsage
-- @usage PluginResourceTools.GetMemUsage(addon)
-- @param addon We'll get the mem usage for this addon
-- @param usestarlibs Use this to change the addon name to StarLibs-1.0 if StarLibs-1.0, the addon, is loaded..
-- @return mem, percent, memdiff, totalmem, totaldiff
function PluginResourceTools.GetMemUsage(addon, userstarlibs)
	assert(type(addon) == "string", "GetCPUUsage: addon name is not a string")	

	if not meminit then return -1, -1, -1, -1, -1 end
	
	if StarLibs and (addon == "StarTip" or addon == "LCD4WoW" or usestarlibs) then
		addon = "StarLibs-1.0"
	end

	if not oldmem[addon] then
		oldmem[addon] = new(-1, -1, -1, -1, -1)
	end
	
	local mem = oldmem[addon][1]
	local perc = oldmem[addon][2]
	local memdiff = oldmem[addon][3]
	local totalmem = oldmem[addon][4]
	local totaldiff = oldmem[addon][5]
	
	return mem, perc, memdiff, totalmem, totaldiff
end


local lastlat = 0
function PluginResourceTools.GetLatency()
	local _, _, latency = GetNetStats()
	if type(latency) == "number" then lastlat = latency else latency = lastlat end
	return latency
end

function PluginResourceTools.GetFramerate()
	return GetFramerate() or 0
end
